/*

TO DO:
 - metric weights => imperial weights
 - surfaces
 - volumes
 - m/s,mps (to both)
 - update example page with speeds plus everything above
 - example page: adaptive results -- in2m, m2in, kmph2miph, miph2kmph, lb2kg
*/

// $Id$

/* ***** BEGIN LICENSE BLOCK ***** 
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Bogdan Stancescu.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * by the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Bogdan Stancescu <converter@moongate.ro>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
  This string defines which unit is parsed by what converter.
  
  Notes:
  1. Since we separate units when parsing the selection by spaces,
     it is safe to also separate them by spaces here. Therefore the format
     of this string is ((unit)(<space>)*(converter)(<space>)*)*
  2. The converters are functions, defined below. In order to avoid name clashes
     between the converter functions and other functions in Mozilla, we
     programatically prepend "converter_" to the name of the converter defined
     here when we evaluate.
  3. Make sure there is a backslash at the end of every line in this string.
     A backslash tells the JavaScript interpreter the line is *continued*
     on the next line (i.e. the lines are simply concatenated in the final
     string, no line breaks result in the end). Please note this behaviour is
     variable in JavaScript depending on the browser; luckily we don't have to
     worry about that in this environment. :)
  4. The selection parser is case-insensitive, so all units in here *MUST* be
     in small letters. For instance "C" as a unit in the string below
     would NOT match "15 C" in the selection, but "c" does.
*/

var conv_table=new String("\
  \"               in2m \
  in               in2m \
  inch             in2m \
  inches           in2m \
  \
  mil              mil2m \
  \
  mi               mi2m \
  mile             mi2m \
  miles            mi2m \
  \
  ft               ft2m \
  foot             ft2m \
  feet             ft2m \
  \'               ft2m \
  \
  yard             yd2m \
  yards            yd2m \
  yd               yd2m \
  \
  ounce            oz2kg \
  oz               oz2kg \
  ounces           oz2kg \
  \
  pound            lb2kg \
  pounds           lb2kg \
  lb               lb2kg \
  lbs              lb2kg \
  \
  m                m2in \
  meter            m2in \
  meters           m2in \
  metre            m2in \
  metres           m2in \
  \
  mm               mm2in \
  millimetre       mm2in \
  millimetres      mm2in \
  millimeter       mm2in \
  millimeters      mm2in \
  \
  km               km2in \
  kilometre        km2in \
  kilometres       km2in \
  kilometer        km2in \
  kilometers       km2in \
  \
  m               microm2in \
  micrometre       microm2in \
  micrometres      microm2in \
  micrometer       microm2in \
  micrometers      microm2in \
  micron           microm2in \
  microns          microm2in \
  \
  cm               cm2in \
  centimetre       cm2in \
  centimetres      cm2in \
  centimeter       cm2in \
  centimeters      cm2in \
  \
  c               c2f \
  c                c2f \
  celsius          c2f \
  \
  f               f2c \
  f                f2c \
  fahrenheit       f2c \
  \
  mph              miph2kmph \
  mi/h             miph2kmph \
  mile/h           miph2kmph \
  miles/h          miph2kmph \
  mile/hour        miph2kmph \
  miles/hour       miph2kmph \
  \
  km/h             kmph2miph \
  kmph             kmph2miph \
  kilometers/hour  kmph2miph \
  kilometres/hour  kmph2miph \
  \
  degree           deg2all \
  degrees          deg2all \
  deg              deg2all \
                  deg2all \
  \
  m/s              mps2all \
  mps              mps2all \
");

/*

  ====================================================================
                          THE CONVERTERS
  ====================================================================
  
  Ok, now that we know which unit is parsed with what converter, let's
  define the converters themselves. If you add extra converters, please
  make sure the order of the functions below is the same as in the string
  above.
  
*/

function converter_in2m(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.0254;
  result.inUnit="in";
  result.outUnit="m";
  return(result);
}

function converter_mil2m(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.0000254;
  result.inUnit="mil";
  result.outUnit="m";
  return(result);
}

function converter_mi2m(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*1609.344;
  result.inUnit="miles";
  result.outUnit="m";
  return(result);
}

function converter_ft2m(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.3048;
  result.inUnit="ft";
  result.outUnit="m";
  return(result);
}

function converter_yd2m(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.9144;
  result.inUnit="yd";
  result.outUnit="m";
  return(result);
}

function converter_oz2kg(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.0283;
  result.inUnit="oz";
  result.outUnit="kg";
  return(result);
}

function converter_lb2kg(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*0.45359237;
  result.inUnit="lbs";
  result.outUnit="kg";
  return(result);
}

function converter_m2in(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v*1000/25.4;
  result.inUnit="m";
  result.outUnit="in";
  return(result);
}

function converter_mm2in(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v/25.4;
  result.inUnit="mm";
  result.outUnit="in";
  return(result);
}

function converter_km2in(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=(v/0.0254)*1000;
  result.inUnit="km";
  result.outUnit="in";
  return(result);
}

function converter_microm2in(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v/25400;
  result.inUnit="m";
  result.outUnit="in";
  return(result);
}

function converter_cm2in(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=v/2.54;
  result.inUnit="cm";
  result.outUnit="in";
  return(result);
}

function converter_c2f(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=Math.round((v*1.8+32)*10)/10;
  result.inUnit="C";
  result.outUnit="F";
  return(result);
}

function converter_f2c(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=Math.round(10*(v-32)/1.8)/10;
  result.inUnit="F";
  result.outUnit="C";
  return(result);
}

function converter_miph2kmph(v) {
  var result=new Object();
  result.inValue=v;
  if (v<2) {
    result.outValue=Math.round(v*16093.44)/10000;
  } else if (v < 10) {
    result.outValue=Math.round(v*160.9344)/100;
  } else {
    result.outValue=Math.round(v*1.609344);
  }
  result.inUnit="mph";
  result.outUnit="km/h";
  return(result);
}

function converter_kmph2miph(v) {
  var result=new Object();
  result.inValue=v;
  if (v<3) {
    result.outValue=Math.round(v/0.0001609344)/10000;
  } else if (v<15) {
    result.outValue=Math.round(v/0.01609344)/100;
  } else {
    result.outValue=Math.round(v/1.609344);
  }
  result.inUnit="km/h";
  result.outUnit="mph";
  return(result);
}

function converter_deg2all(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=0;
  var tmp1=Math.round((v*1.8+32)*10)/10; // C=>F
  var tmp2=Math.round(10*(v-32)/1.8)/10; // F=>C
  result.outCustom="> "+v+"C = "+tmp1+"F; "+v+"F = "+tmp2+"C";
  // Both lines below as to not match anything
  result.inUnit="";
  result.outUnit="";
  return(result);
}

function converter_mps2all(v) {
  var result=new Object();
  result.inValue=v;
  result.outValue=0;
  var tmp1=Math.round(v * 1800 / 5)/100; // m/s => km/h
  var tmp2=Math.round(v * 3600 / 1610.3*100)/100; // m/s=>mph
  result.outCustom=tmp1+" km/h ("+tmp2+" mph)";
  // Both lines below as to not match anything
  result.inUnit="m/s";
  result.outUnit="";
  return(result);
}

/*

  ====================================================================
                              ADJUSTERS
  ====================================================================

  These functions perform order of magnitude adjustments within the same
  units system, trying to guess the most intuitive unit. For instance
  it would be counter-intuitive to show a distance of the order of
  hundreds of kilometers in millimeters.
  
*/

function metricWeightAdjustments(o) {
  if (o.outUnit!="kg") {
    return(o);
  }
  var tv=o.outValue;
  if ((tv>=1) && (tv<1000)) {
    return(o);
  }
  if (tv>=1000) {
    o.outValue/=1000;
    o.outUnit="t";
    return(o);
  }
  if (tv>=0.001) {
    o.outValue*=1000;
    o.outUnit="g";
    return(o);
  }
  o.outValue*=1000000;
  o.outUnit="mg";
  return(o);
}

// I'm not familiar enough with the imperial units, these may not be the most
// intuitive choices; I hope they're acceptable though -- if not, you're free
// to send me a patch any time! :)
function imperialDistanceAdjustments(o) {
  if (o.outUnit!="in") {
    return(o);
  }
  var tv=o.outValue;
  if (tv<(1.0/32.0)) {
    o.outValue=tv*100;
    o.outUnit="mil";
    return(o);
  }
  if (tv>=5280*36) {
    o.outValue=Math.round(tv/12/52.8)/100;
    o.outUnit="mi";
    return(o);
  }
  if (tv>=36000) {
    o.outCustom=(Math.round(tv/0.36)/100)+" yd ("+Math.round(tv/12/52.8)/100+" mi)";
    return(o);
  }
  if (tv>=36) {
    o.outCustom=(Math.round(tv/0.36)/100)+" yd ("+(Math.round(tv/0.12)/100)+" ft)";
    return(o);
  }
  var tv32=Math.round(tv*32) % 32;
  if (tv32==16) {
    var fraction="1/2";
  } else if (!tv32 % 8) {
    var fraction=(tv32/8)+"/4";
  } else if (!tv32 % 4) {
    var fraction=(tv32/4)+"/8";
  } else if (!tv32 % 2) {
    var fraction=(tv32/2)+"/16";
  } else {
    var fraction=tv32+"/32";
  }
  tv=Math.round(tv);
  var inches=tv % 12;
  tv=(tv-inches)/12;
  o.outCustom=tv+"'"+inches+"-"+fraction+"\"";
  return(o);
}

function metricDistanceAdjustments(o) {
  if (o.outUnit!="m") {
    return(o);
  }
  var tv=o.outValue;
  if ((tv>=1) && (tv<1000)) {
    return(o);
  }
  if (tv>=1000) {
    o.outValue/=1000;
    o.outUnit="km";
    return(o);
  }
  if (tv>=0.01) {
    o.outValue*=100;
    o.outUnit="cm";
    return(o);
  }
  if (tv>=0.0001) {
    o.outValue*=1000;
    o.outUnit="mm";
    return(o);
  }
  o.outValue*=1000000;
  o.outUnit="m";
  return(o);
}

/*

  ____________________________________________________________________

   Typically you shouldn't need to change anything below this line.
  ____________________________________________________________________



  ====================================================================
                           REGISTRATION
  ====================================================================
  
  This is the Mozilla internal stuff, where we register the converter
  entry in the popup. Traditionally this stays at the beginning of the
  script, but I preferred to leave the converters at the top so it's as
  easy as possible for future developers to write new converters.
  
  If you need to write new adjusters, you'll have to register them
  as an if case within converterPopup(). Use the existing adjusters
  as a model.

*/

window.addEventListener("load",converterInit,false);

var converterBundle;
function converterInit() {
  converterBundle=document.getElementById("converter_string_bundle");
  var tmpElement=document.getElementById("contentAreaContextMenu");
  if (tmpElement)
    tmpElement.addEventListener("popupshowing",onConverterPopup,false);
}

/**
* This function is executed when the popup menu pops up. This function is the
* brain of this script, it coordinates all the logic around.
* Too much to explain in words, this script is anyway overdocumented, take
* a look at the code and you'll understand the logic once you know what each
* function does.
*/
function onConverterPopup() {
  var item = document.getElementById("context_converterselect");
  if (!gContextMenu.isTextSelected){
    item.hidden = true;
    return(true);
  }
  var temp = converterGetSelectedWord();
  if (!temp) {
    item.hidden = true;
    return(true);
  }
  var conversions = registerConverters();
  var pos=inArray(temp[1],conversions[0]);
  if (pos==-1) {
    item.hidden = true;
    return(true);
  }
  temp[0]=determine_separators(temp[0]);
  var converter=conversions[1][pos];
  var cr=eval("converter_"+converter+"("+temp[0]+")"); // conversion result
  if (cr.outUnit=="m") {
    cr=metricDistanceAdjustments(cr);
  }
  if (cr.outUnit=="kg") {
    cr=metricWeightAdjustments(cr);
  }
  if (cr.outUnit=="in") {
    cr=imperialDistanceAdjustments(cr);
  }
  if (!cr.outCustom) {
    cr.outCustom=cr.outValue+" "+cr.outUnit;
  }
  item.hidden = false;
  var label = cr.inValue + " " + cr.inUnit + " = " + cr.outCustom;
  item.setAttribute("label", label);
}




/*

  ====================================================================
                         INTERNAL CONVERTER CODE
  ====================================================================
  
  These functions are internal to the converter itself, they perform
  various tasks documented above each function.
  
*/

/**
* This function receives a string which presumably contains a number.
* The number may be formatted with either comma or period as decimal separator,
* and the other one as thousands separator. This function tries to determine
* which way it is, defaulting to English format if ambiguous.
* It returns a proper number.
*/
function determine_separators(s) {
  // We need to get a string, otherwise we don't care at all
  if (!isNaN(s)) {
    return(s);
  }
  // Ok, first we'll take the default out of the way
  if (s.indexOf(",")==-1) {
    s=s.replace(/\./g,"");
    return(s);
  }
  // Now we're sure we have a comma in there somewhere. Let's dissect this:
  if (s.indexOf(".")!=-1) {
    // Good, we also have a period, this makes it easy to determine which case
    // it is: the last one is the decimal separator.
    if (s.lastIndexOf(".")>s.lastIndexOf(",")) {
      // If it's the period, we just remove the commas and return the string
      s=s.replace(/,/g,"");
    } else {
      // If it's the comma, we remove the periods, change the comma into a
      // period and return the string.
      s=s.replace(/\./g,"");
      s=s.replace(/,/g,".");
    }
  } else {
    // Ewww, this is nasty, we positively have a comma, but we have no
    // period to serve as a witness. We'll branch in two once again:
    if (s.indexOf(",")!=s.lastIndexOf(",")) {
      // Either there's more than one comma, in which case the period would
      // be the decimal separator...
      s=s.replace(/,/g,"");
    } else {
      // ...or there's a single comma. If there's only one comma, then the
      // number has only ony chance to be suspected of using comma as a
      // thousands separator: it MUST be in the proper position for it to be a
      // thousands separator AND the number must be no larger than 100,000.
      // In which case we'll ASSUME the default case (period is assumed to be
      // the decimal separator). This is not perfect, but it's the safest bet.
      // If either of the conditions above is not met, we'll use the comma
      // as decimal separator.
      if ((s.length<=7) && (s.indexOf(",")==s.length-4)) {
        s=s.replace(/,/g,"");
      } else {
        s=s.replace(/\./g,"");
        s=s.replace(/,/g,".");
      }
    }
  }
  return(s);
}

/**
* This function is called when the user clicks on the converter entry
* in the popup.
*/
function onConvert(){
  // Doing nothing at the moment, might pop up some window or copy the
  // result of the conversion to clipboard, we'll see about that.
}

/*
* This function registers the converters defined at the beginning of this
* file. What it basically does is split that string into items and builds
* two arrays, on of which contains the units and the other the converters.
* It then builds an array with two elements which are the arrays previously
* described, and returns it. It's used by onConverterPopup().
*/
function registerConverters() {
  var conv_symbols=new Array();
  var conv_functions=new Array();
  var atoms=conv_table.split(" ");
  var conv_symbol=new Boolean(true);
  for(i=0;i<atoms.length;i++) {
    atom=atoms[i];
    if (!atom) {
      continue;
    }
    if (conv_symbol) {
      conv_symbols[conv_symbols.length]=atom;
    } else {
      conv_functions[conv_functions.length]=atom;
    }
    conv_symbol=!conv_symbol;
  }
  var result=new Array(conv_symbols,conv_functions);
  return(result);
}

/**
* A simple function to emulate indexOf() for arrays. Returns the position
* of the needle in the haystack array, or -1 if not found. Used by
* onConverterPopup() to determine the units in the selection by matching
* the 2nd result of converterGetSelectedWord() below against the second array
* returned by registerConverters() above.
*/
function inArray(needle, haystack) {
  for(i=0;i<haystack.length;i++) {
    var hayAtom=haystack[i];
    if(needle==hayAtom) {
      return(i);
    }
  }
  return(-1);
}

/**
* This function parses the current selection, searching for the first
* substring which looks like a dimension. The typical dimension understood
* by this function is a substring made of a number, optionally followed by
* one or more spaces, followed by a word (loose definition here, check
* the source for the actual regexp). Before looking for the typical dimension
* however, it must look for the awkward imperial formats in
* feet/inches/fractions because otherwise it could end up only matching the
* feet in such a format.
*
* It returns an array made of two elements: the first one is the number and
* the second one is the unit.
*
* Please note that for typical measurements it also checks for fractions,
* and if found it tries to determine the decimal separator for each term and
* evaluates them numerically. The first element of the resulting array is
* therefore not guaranteed to be a string--in this case, it's a number.
*/
function converterGetSelectedWord(){
  var word = gContextMenu.searchSelected();
  
  // Let's start by checking for the awkward Imperial notations.
  
  // First checking for egineering format (1'2 1/32" or 1'2-1/32").
  // If found, we convert the whole thing to inches and return as if we
  // encountered the corresponding string in inches using decimal notation
  // instead of fractions.
  re = /([0-9]*)\'([0-9]+)[ \-]([0-9\/]+)/;
  var result = re.exec(word);
  if (result) {
    var Aresult=new Array();
    Aresult[0]=12*result[1]+1*result[2]+eval(result[3]);
    Aresult[1]="in";
    return(Aresult);
  }
  
  // And finally checking for informal format (5'10)
  re = /([0-9]+)\'([0-9]{1,2})/;
  var result = re.exec(word);
  if (result) {
    var Aresult=new Array();
    Aresult[0]=12*result[1]+1*result[2];
    Aresult[1]="in";
    return(Aresult);
  }

  // Ok, didn't find any of the wicked ones, let's check for
  // some reasonable measurements AND fractions.
  re = /([\-]?[0-9][0-9.,\/]*)[\s]*([^ .,!?;\-+\(\)]+)/i;
  result = re.exec(word);
  // Must have the first (result), or the whole thing blows for null results
  if ((result) && (result[1]) && (result[2])) {
    var Aresult=new Array();
    var tmpTest=new String(result[1]);
    if (result[1].indexOf("/")!=-1) {
      var resultMembers=result[1].split("/");
      for(i=0;i<resultMembers.length;i++) {
        resultMembers[i]=determine_separators(resultMembers[i]);
      }
      result[1]=eval(resultMembers.join("/"));
    }
    Aresult[0]=result[1];
    Aresult[1]=result[2].toLowerCase();
    return(Aresult);
  }
  
  // Yikes, you're not using the converter at all, you naughty, naughty boy!
  return(false);
}
